// Author: Kurt Schwehr
// Copyright 2015 Google Inc. All Rights Reserved.
// License: Match GDAL or Apache 2.0.
#include <limits>
#include <algorithm>
#include <utility>
#include <iostream>

#if 0
bool AddWillOverflowOld(unsigned short a, unsigned short b) {
  if (b > a) {
    std::swap(a, b);
  }
  if ( ( std::numeric_limits<unsigned short>::max() - a ) < b )
    return true;
  return false;
}
#endif

template<typename T>
bool CPLAddWillOverflow(T a, T b) {
  if (a > b) {
    if ( ( std::numeric_limits<T>::max() - a ) < b )
      return true;
    return false;
  }
  if ( ( std::numeric_limits<T>::max() - b ) < a )
    return true;
  return false;
}

int main() {
  {
    unsigned short a = 0xFFFD;
    unsigned short b = 0x0000;
    unsigned short c = 0x0001;
    unsigned short d = 0x0002;
    unsigned short e = 0x0003;

    std::cout << a << " + " << b << " => " << (CPLAddWillOverflow(a, b) ? "overflow" : "okay") << "\n";
    std::cout << a << " + " << c << " => " << (CPLAddWillOverflow(a, c) ? "overflow" : "okay") << "\n";
    std::cout << a << " + " << d << " => " << (CPLAddWillOverflow(a, d) ? "overflow" : "okay") << "\n";
    std::cout << a << " + " << e << " => " << (CPLAddWillOverflow(a, e) ? "overflow" : "okay") << "\n";
    std::cout << b << " + " << a << " => " << (CPLAddWillOverflow(b, a) ? "overflow" : "okay") << "\n";
    std::cout << c << " + " << a << " => " << (CPLAddWillOverflow(c, a) ? "overflow" : "okay") << "\n";
    std::cout << d << " + " << a << " => " << (CPLAddWillOverflow(d, a) ? "overflow" : "okay") << "\n";
    std::cout << e << " + " << a << " => " << (CPLAddWillOverflow(e, a) ? "overflow" : "okay") << "\n";
  }

  {
    short a = 0x7FFD;
    short b = 0x0000;
    short c = 0x0001;
    short d = 0x0002;
    short e = 0x0003;

    std::cout << a << " + " << b << " => " << (CPLAddWillOverflow(a, b) ? "overflow" : "okay") << "\n";
    std::cout << a << " + " << c << " => " << (CPLAddWillOverflow(a, c) ? "overflow" : "okay") << "\n";
    std::cout << a << " + " << d << " => " << (CPLAddWillOverflow(a, d) ? "overflow" : "okay") << "\n";
    std::cout << a << " + " << e << " => " << (CPLAddWillOverflow(a, e) ? "overflow" : "okay") << "\n";
    std::cout << b << " + " << a << " => " << (CPLAddWillOverflow(b, a) ? "overflow" : "okay") << "\n";
    std::cout << c << " + " << a << " => " << (CPLAddWillOverflow(c, a) ? "overflow" : "okay") << "\n";
    std::cout << d << " + " << a << " => " << (CPLAddWillOverflow(d, a) ? "overflow" : "okay") << "\n";
    std::cout << e << " + " << a << " => " << (CPLAddWillOverflow(e, a) ? "overflow" : "okay") << "\n";
  }

  return 0;
}

/* Example test output:

g++ -std=c++98 -o integer_overflow -g -Wall -Weverything integer_overflow.cc
# or
g++ -std=c++14 -o integer_overflow -g -Wall -Weverything integer_overflow.cc

./integer_overflow
65533 + 0 => okay
65533 + 1 => okay
65533 + 2 => okay
65533 + 3 => overflow
0 + 65533 => okay
1 + 65533 => okay
2 + 65533 => okay
3 + 65533 => overflow
32765 + 0 => okay
32765 + 1 => okay
32765 + 2 => okay
32765 + 3 => overflow
0 + 32765 => okay
1 + 32765 => okay
2 + 32765 => okay
3 + 32765 => overflow

 */
